import std/[logging, options, os, strutils]
import zero_functional

import ./ast

type
  FormatContext = ref object
    chart: ChordChart
    currentKey: ChordChartPitch
    sourceKey: ChordChartPitch
    currentSection: ChordChartNode

const DEFAULT_STYLESHEET* = """
<style>
* {
  box-sizing: border-box;
  margin: 0;
  padding: 0;
}

html { font-family: sans-serif; }

.page-contents {
  column-width: 336px;
  column-width: 3.5in;
}

.column-break { margin-bottom: auto; }

h2 { font-size: 1.25em; }

section {
  break-inside: avoid;
  padding-top: 1em;
}

h3 {
  font-size: 1.125rem;
  margin-bottom: 0.5em;
  text-decoration: underline;
}

.line {
  display: flex;
  flex-direction: row;
  align-items: flex-end;

  margin-left: 0.5em;
  margin-bottom: 0.5em;
}

.word {
  display: flex;
  flex-direction: column;
  align-items: flex-start;
}

.word.space-after { margin-right: 0.5em; }

.chord {
  font-weight: 600;
  margin-right: 0.5em;
}

.song-order h3 {
  font-style: italic;
  font-weight: normal;
}

.song-order li {
  list-style: none;
  margin-left: 1em;
}

@media screen {
  body { margin: 1em; }
}

@media print {
  .page-contents { column-count: 2; }
}
</style>
"""

proc transpose(ctx: FormatContext, chord: ChordChartChord): string =
  let distance = ctx.currentKey - ctx.sourceKey

  if distance != 0:
    result = $(chord.rootPitch + distance)
    if chord.flavor.isSome: result &= chord.flavor.get
    if chord.bassPitch.isSome: result &= "/" & $(chord.bassPitch.get + distance)
    #debug "transposed " & $ctx.sourceKey & " -> " & $ctx.currentKey &
    #  " (distance " & $distance & ")\p\tchord: " & $chord & " to " & result

  else: result = chord.original

func makeSongOrder(songOrder, indent: string): string =
  result = indent & "<section class=song-order>\p" &
           indent & "  <h3>Song Order</h3>\p" &
           indent & "  <ul>\p"
  result &= join(songOrder.split(",") -->
    map(indent & "    <li>" & it.strip & "</li>\p"), "")
  result &= indent & "  </ul>\p" & indent & "</section>\p"

proc toHtml(ctx: var FormatContext, node: ChordChartNode, indent: string): string =
  result = ""
  case node.kind
  of ccnkSection:
    ctx.currentSection = node
    result &= indent & "<section>\p" &
      indent & "  " & "<h3>" & node.sectionName & "</h3>\p"

    var contents = newSeq[string]()
    for contentNode in node.sectionContents:
      contents.add(ctx.toHtml(contentNode, indent & "  "))
    result &= contents.join("\p")

    result &= indent & "</section>"
    ctx.currentSection = EMPTY_CHORD_CHART_NODE

  of ccnkLine:
    result &= indent & "<div class=line>\p"
    for linePart in node.line: result &= ctx.toHtml(linePart, indent & "  ")
    result &= indent & "</div>"

  of ccnkWord:
    result &= "<span class='word"
    #if node.spaceBefore: result&=" space-before"
    if node.spaceAfter: result&=" space-after"
    result &= "'>"

    if node.chord.isSome:
      result &= "<span class=chord>" & ctx.transpose(node.chord.get) & "</span>"

    result &= "<span class=lyric>"
    if node.word.isSome: result &= node.word.get
    else: result &= "&nbsp;"
    result &= "</span>"

    result &= "</span>"

  of ccnkNote:
    result &= indent & "<div class=note>" & node.note & "</div>"

  of ccnkColBreak:
    result &= "<div class=column-break />"

  of ccnkPageBreak:
    result &= "</div><div class=page-contents>"

  of ccnkTransposeKey:
    ctx.currentKey = ctx.currentKey + node.transposeSteps
    let headingVal = indent & "<h4 class=note>Key Change: " & $ctx.currentKey & "</h4>"
    if ctx.currentSection.kind == ccnkNone: result &= headingVal
    else:
      result &= "</section><section>" & headingVal & "</section><section>"

  of ccnkRedefineKey:
    let oldKey = ctx.sourceKey
    ctx.sourceKey = node.newKey
    ctx.currentKey = ctx.sourceKey

    if oldKey != ctx.currentKey:
      let headingVal = indent & "<h4 class=note>Key Change: " & $ctx.currentKey & "</h4>"
      if ctx.currentSection.kind == ccnkNone: result &= headingVal
      else:
        result &= "</section><section>" & headingVal & "</section><section>"

  of ccnkNone: discard

proc toHtml*(
    cc: ChordChart,
    stylesheets = @[DEFAULT_STYLESHEET],
    scripts: seq[string] = @[]): string =

  var ctx = FormatContext(
    chart: cc,
    currentKey: cc.metadata.key,
    sourceKey: cc.metadata.key,
    currentSection: EMPTY_CHORD_CHART_NODE)

  result = """<!doctype html>
<html>
  <head>
    <title>""" & cc.metadata.title & "</title>\p"

  for ss in stylesheets:
    if ss.startsWith("<style>"): result &= ss
    elif fileExists(ss): result &= "<style>\p" & readFile(ss) & "\p</style>"
    else: warn "cannot read stylesheet file '" & ss & "'"

  for sc in scripts:
    if sc.startsWith("<script"): result &= sc
    elif fileExists(sc): result &= "<script>\p" & readFile(sc) & "\p</script>"
    else: warn "cannot read script file '" & sc & "'"

  result &= "  </head>\p  <body>"

  var indent = "      "
  #
  # Title
  result &= indent & "<h1>" & cc.metadata.title & "</h1>\p"

  var metadataPieces = @["Key: " & $cc.metadata.key]
  if cc.metadata.contains("time signature"):
    metadataPieces.add(cc.metadata["time signature"])
  if cc.metadata.contains("bpm"):
    metadataPieces.add(cc.metadata["bpm"] & "bpm")
  if cc.metadata.contains("tempo"):
    metadataPieces.add(cc.metadata["tempo"])

  result &= indent & "<h2>" & metadataPieces.join(" | ") & "</h2>\p"

  result &= "<div class=page-contents>"
  result &= join(cc.nodes --> map(ctx.toHtml(it, indent & "  ")), "\p")

  if cc.metadata.contains("song order"):
    result &= makeSongOrder(cc.metadata["song order"], indent)

  result &= "</div>"

  result &= "  </body>\p</html>"
